#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: MIT
# See https://github.com/aboutcode-org/django-altcha for support or download.
# See https://aboutcode.org for more information about AboutCode FOSS projects.
#

from datetime import datetime
from unittest import mock

from django.test import TestCase

from django_altcha import get_altcha_challenge

mock_now = datetime(2025, 10, 10)


class DjangoAltchaUtilsTest(TestCase):
    def test_get_altcha_challenge_max_number(self):
        challenge = get_altcha_challenge()
        self.assertEqual(1000000, challenge.max_number)
        challenge = get_altcha_challenge(max_number=50)
        self.assertEqual(50, challenge.max_number)

    @mock.patch("django_altcha.datetime.datetime")
    def test_get_altcha_challenge_expire(self, mock_datetime):
        mock_datetime.now.return_value = mock_now
        mock_datetime.side_effect = lambda *args, **kwargs: datetime(*args, **kwargs)

        # Default ALTCHA_CHALLENGE_EXPIRE is applied
        challenge = get_altcha_challenge()
        salt_expires = challenge.salt.split("?expires=")[-1]
        self.assertEqual("1760073600", salt_expires)

        # Provided `expires` argument is applied
        challenge = get_altcha_challenge(expires=10000)
        salt_expires = challenge.salt.split("?expires=")[-1]
        self.assertEqual("1760072410", salt_expires)

        # Custom ALTCHA_CHALLENGE_EXPIRE value is applied
        with mock.patch("django_altcha.ALTCHA_CHALLENGE_EXPIRE", 9999):
            challenge = get_altcha_challenge()
            salt_expires = challenge.salt.split("?expires=")[-1]
            self.assertEqual("1760072409", salt_expires)



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: MIT
# See https://github.com/aboutcode-org/django-altcha for support or download.
# See https://aboutcode.org for more information about AboutCode FOSS projects.
#

import time
import unittest
from unittest import mock

from django.core.cache.backends.locmem import LocMemCache
from django.test import override_settings

from django_altcha import ALTCHA_CHALLENGE_EXPIRE_SECONDS
from django_altcha import get_altcha_cache
from django_altcha import is_challenge_used
from django_altcha import mark_challenge_used


class DjangoAltchaCacheTest(unittest.TestCase):
    def setUp(self):
        self.challenge = "test-challenge-123"

    @override_settings(ALTCHA_CACHE_ALIAS="altcha")
    @mock.patch("django_altcha.caches")
    def test_get_altcha_cache_with_alias(self, mock_caches):
        get_altcha_cache()
        mock_caches.__getitem__.assert_called_once_with("altcha")

    @override_settings(ALTCHA_CACHE_ALIAS=None)
    def test_get_altcha_cache_without_alias_uses_locmemcache(self):
        cache = get_altcha_cache()
        self.assertIsInstance(cache, LocMemCache)
        self.assertEqual(cache.default_timeout, ALTCHA_CHALLENGE_EXPIRE_SECONDS)

    def test_mark_and_check_challenge_used(self):
        cache = get_altcha_cache()
        cache.clear()

        self.assertFalse(is_challenge_used(self.challenge))
        mark_challenge_used(self.challenge, timeout=ALTCHA_CHALLENGE_EXPIRE_SECONDS)
        self.assertTrue(is_challenge_used(self.challenge))

    def test_challenge_expires(self):
        cache = get_altcha_cache()
        cache.clear()

        mark_challenge_used(self.challenge, timeout=1)  # 1 second timeout
        self.assertTrue(is_challenge_used(self.challenge))
        time.sleep(1.1)
        self.assertFalse(is_challenge_used(self.challenge))



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: MIT
# See https://github.com/aboutcode-org/django-altcha for support or download.
# See https://aboutcode.org for more information about AboutCode FOSS projects.
#

from django.test import TestCase
from django.urls import reverse


class DjangoAltchaViewTest(TestCase):
    def test_challenge_view_returns_200(self):
        response = self.client.get(reverse("altcha_challenge"))
        self.assertEqual(response.status_code, 200)

    def test_challenge_view_returns_json(self):
        response = self.client.get(reverse("altcha_challenge"))
        self.assertEqual(response["Content-Type"], "application/json")

    def test_challenge_response_contains_expected_keys(self):
        response = self.client.get(reverse("altcha_challenge"))
        data = response.json()

        expected_keys = ["algorithm", "challenge", "max_number", "salt", "signature"]
        self.assertEqual(expected_keys, list(data.keys()))

        self.assertEqual("SHA-256", data["algorithm"])
        self.assertEqual(100, data["max_number"])



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: MIT
# See https://github.com/aboutcode-org/django-altcha for support or download.
# See https://aboutcode.org for more information about AboutCode FOSS projects.
#

import base64
import json
from unittest import mock

from django import forms
from django.test import TestCase
from django.test import override_settings

from django_altcha import ALTCHA_CHALLENGE_EXPIRE
from django_altcha import AltchaField
from django_altcha import AltchaWidget
from django_altcha import is_challenge_used

TEST_CHALLENGE = "test-challenge-123"


def make_valid_payload(challenge=TEST_CHALLENGE):
    payload_dict = {"challenge": challenge}
    json_str = json.dumps(payload_dict)
    encoded_bytes = base64.b64encode(json_str.encode("utf-8"))
    return encoded_bytes.decode("utf-8")


class DjangoAltchaFieldTest(TestCase):
    def setUp(self):
        class TestForm(forms.Form):
            altcha_field = AltchaField()

        self.form_class = TestForm

    def test_altcha_field_renders_widget(self):
        form = self.form_class()
        self.assertIsInstance(form.fields["altcha_field"].widget, AltchaWidget)

    def test_altcha_field_widget_default_options(self):
        altcha_field = AltchaField()
        self.assertEqual(ALTCHA_CHALLENGE_EXPIRE, altcha_field.widget.options["expire"])
        altcha_field = AltchaField(expire=1)
        self.assertEqual(1, altcha_field.widget.options["expire"])
        altcha_field = AltchaField(expire=None)
        self.assertEqual(None, altcha_field.widget.options["expire"])

    def test_altcha_field_options_to_widget(self):
        altcha_field = AltchaField(maxnumber=50, expire=10000)
        self.assertEqual(50, altcha_field.widget.options["maxnumber"])
        self.assertEqual(10000, altcha_field.widget.options["expire"])

    def test_altcha_field_with_missing_value_raises_required_error(self):
        form = self.form_class(data={})
        self.assertFalse(form.is_valid())
        self.assertIn("altcha_field", form.errors)
        self.assertEqual(
            form.errors["altcha_field"][0], "ALTCHA CAPTCHA token is missing."
        )

    @override_settings(ALTCHA_CACHE_ALIAS=None)
    @mock.patch("altcha.verify_solution")
    def test_altcha_field_validation_calls_verify_solution(self, mock_verify_solution):
        self.assertFalse(is_challenge_used(TEST_CHALLENGE))
        mock_verify_solution.return_value = (True, None)
        valid_payload = make_valid_payload()
        form = self.form_class(data={"altcha_field": valid_payload})
        self.assertTrue(form.is_valid())
        mock_verify_solution.assert_called_once_with(
            payload=valid_payload,
            hmac_key=mock.ANY,
            check_expires=True,
        )

        # Replay the validation using the same challenge
        self.assertTrue(is_challenge_used(TEST_CHALLENGE))
        form = self.form_class(data={"altcha_field": valid_payload})
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors["altcha_field"][0], "Challenge has already been used."
        )

    @mock.patch("altcha.verify_solution")
    def test_altcha_field_validation_fails_on_invalid_token(self, mock_verify_solution):
        mock_verify_solution.return_value = (False, "Invalid token")
        form = self.form_class(data={"altcha_field": "invalid_token"})
        self.assertFalse(form.is_valid())
        self.assertIn("altcha_field", form.errors)
        self.assertEqual(form.errors["altcha_field"][0], "Invalid CAPTCHA token.")

    @mock.patch("altcha.verify_solution")
    def test_altcha_field_validation_handles_exception(self, mock_verify_solution):
        mock_verify_solution.side_effect = Exception("Verification failed")
        form = self.form_class(data={"altcha_field": "some_token"})
        self.assertFalse(form.is_valid())
        self.assertIn("altcha_field", form.errors)
        self.assertEqual(
            form.errors["altcha_field"][0], "Failed to process CAPTCHA token"
        )



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: MIT
# See https://github.com/aboutcode-org/django-altcha for support or download.
# See https://aboutcode.org for more information about AboutCode FOSS projects.
#

import base64
import datetime
import json
import secrets

from django import forms
from django.conf import settings
from django.core.cache import caches
from django.core.cache.backends.locmem import LocMemCache
from django.forms.widgets import HiddenInput
from django.http import JsonResponse
from django.utils.decorators import method_decorator
from django.utils.translation import gettext_lazy as _
from django.views import View
from django.views.decorators.http import require_GET

import altcha

__version__ = "0.2.0"
VERSION = __version__

# Get the ALTCHA_HMAC_KEY from the settings, or generate one if not present.
ALTCHA_HMAC_KEY = getattr(settings, "ALTCHA_HMAC_KEY", secrets.token_hex(32))
ALTCHA_JS_URL = getattr(settings, "ALTCHA_JS_URL", "/static/altcha/altcha.min.js")

# Challenge expiration duration in milliseconds.
# Default to 20 minutes as per Altcha security recommendations.
# https://altcha.org/docs/v2/security-recommendations/
ALTCHA_CHALLENGE_EXPIRE = getattr(settings, "ALTCHA_CHALLENGE_EXPIRE", 1200000)
ALTCHA_CHALLENGE_EXPIRE_SECONDS = ALTCHA_CHALLENGE_EXPIRE // 1000


def get_altcha_cache():
    """
    Returns a Django cache backend instance to be used for storing ALTCHA challenge
    data, especially for replay attack protection.

    - If the setting `ALTCHA_CACHE_ALIAS` is set, the cache with that alias
      will be used.
    - If not set, a local in-memory cache will be used with a timeout matching
      the challenge expiration in seconds.
    """
    cache_alias = getattr(settings, "ALTCHA_CACHE_ALIAS", None)
    if cache_alias:
        return caches[cache_alias]

    # Use the same timeout for the cache as the challenge expiration to ensure
    # cached challenges expire in sync with their validity period.
    params = {"timeout": ALTCHA_CHALLENGE_EXPIRE_SECONDS}
    return LocMemCache(name="altcha_local", params=params)


_altcha_cache = get_altcha_cache()


def is_challenge_used(challenge):
    """Check if a challenge has already been used."""
    return _altcha_cache.get(key=challenge) is not None


def mark_challenge_used(challenge, timeout):
    """Mark a challenge as used by storing it in the cache with a timeout."""
    _altcha_cache.set(key=challenge, value=True, timeout=timeout)


def get_altcha_challenge(max_number=None, expires=None):
    """
    Generate and return an ALTCHA challenge.

    Attributes:
        max_number (int): Maximum number to use for the challenge.
        expires (int): Expiration time for the challenge in milliseconds.

    Returns:
        altcha.Challenge: The generated challenge.
    """
    expires = expires or ALTCHA_CHALLENGE_EXPIRE
    options = {
        "hmac_key": ALTCHA_HMAC_KEY,
        "expires": datetime.datetime.now() + datetime.timedelta(milliseconds=expires),
    }

    if max_number is not None:
        options["max_number"] = max_number

    challenge = altcha.create_challenge(altcha.ChallengeOptions(**options))
    return challenge


class AltchaWidget(HiddenInput):
    template_name = "altcha_widget.html"

    def __init__(self, options, *args, **kwargs):
        """Initialize the ALTCHA widget with provided options from the field."""
        self.options = options or {}
        super().__init__(*args, **kwargs)

    def get_context(self, name, value, attrs):
        """Generate the widget context, including ALTCHA JS URL and challenge."""
        context = super().get_context(name, value, attrs)
        context["js_src_url"] = ALTCHA_JS_URL

        # If a `challengeurl` is provided, the challenge will be fetched from this URL.
        # This can be a local Django view or an external API endpoint.
        # If not provided, a unique challenge is generated locally in a self-hosted
        # mode.
        # Since the challenge must be fresh for each form rendering, it is generated
        # inside `get_context`, not `__init__`.
        if not self.options.get("challengeurl"):
            challenge = get_altcha_challenge(
                max_number=self.options.get("maxnumber"),
                expires=self.options.get("expire"),
            )
            self.options["challengejson"] = json.dumps(challenge.__dict__)

        context["widget"]["altcha_options"] = self.options
        return context


class AltchaField(forms.Field):
    widget = AltchaWidget
    default_error_messages = {
        "error": _("Failed to process CAPTCHA token"),
        "invalid": _("Invalid CAPTCHA token."),
        "required": _("ALTCHA CAPTCHA token is missing."),
        "replay": _("Challenge has already been used."),
    }
    default_options = {
        # URL of your server to fetch the challenge from.
        "challengeurl": None,
        # JSON-encoded challenge data
        # (use instead of challengeurl to avoid HTTP request).
        "challengejson": None,
        # Automatically verify without user interaction.
        # Possible values: "off", "onfocus", "onload", "onsubmit".
        "auto": None,
        # Artificial delay before verification (in milliseconds, default: 0).
        "delay": None,
        # Challenge expiration duration (in milliseconds).
        "expire": ALTCHA_CHALLENGE_EXPIRE,
        # Enable floating UI.
        # Possible values: "auto", "top", "bottom".
        "floating": None,
        # CSS selector of the “anchor” to which the floating UI is attached.
        # Default: submit button in the related form.
        "floatinganchor": None,
        # Y offset from the anchor element for the floating UI (in pixels, default: 12).
        "floatingoffset": None,
        # Enable a “persistent” mode to keep the widget visible under specific
        # conditions.
        # Possible values: "true", "focus".
        "floatingpersist": None,
        # Hide the footer (ALTCHA link).
        "hidefooter": None,
        # Hide the ALTCHA logo.
        "hidelogo": None,
        # Max number to iterate to (default: 1,000,000).
        "maxnumber": None,
        # JSON-encoded translation strings for customization.
        "strings": None,
        # Automatically re-fetch and re-validate when the challenge expires
        # (default: true).
        "refetchonexpire": None,
        # Number of workers for Proof of Work (PoW).
        # Default: navigator.hardwareConcurrency or 8 (max value: 16).
        "workers": None,
        # URL of the Worker script (default: ./worker.js, only for external builds).
        "workerurl": None,
        # Print log messages in the console (for debugging).
        "debug": None,
        # Causes verification to always fail with a "mock" error.
        "mockerror": None,
        # Generates a “mock” challenge within the widget, bypassing the request to
        # challengeurl.
        "test": None,
    }

    def __init__(self, *args, **kwargs):
        """Initialize the ALTCHA field and pass widget options for rendering."""
        widget_options = {
            key: kwargs.pop(key, self.default_options[key])
            for key in self.default_options
        }
        kwargs["widget"] = self.widget(options=widget_options)
        super().__init__(*args, **kwargs)

    def validate(self, value):
        """Validate the CAPTCHA token and verify its authenticity."""
        super().validate(value)

        if not value:
            raise forms.ValidationError(
                self.error_messages["required"], code="required"
            )

        try:
            verified, error = altcha.verify_solution(
                payload=value,
                hmac_key=ALTCHA_HMAC_KEY,
                check_expires=True,
            )
        except Exception:
            raise forms.ValidationError(self.error_messages["error"], code="error")

        if not verified:
            raise forms.ValidationError(self.error_messages["invalid"], code="invalid")

        self.replay_attack_protection(payload=value)

    def replay_attack_protection(self, payload):
        """Protect against replay attacks by ensuring each challenge is single-use."""
        try:
            # Decode payload from base64 and parse JSON to extract the challenge
            payload_data = json.loads(base64.b64decode(payload).decode())
            challenge = payload_data["challenge"]
        except Exception:
            raise forms.ValidationError(self.error_messages["error"], code="error")

        if is_challenge_used(challenge):
            raise forms.ValidationError(self.error_messages["replay"], code="invalid")

        # Mark as used for the same duration as challenge expiration
        mark_challenge_used(challenge, timeout=ALTCHA_CHALLENGE_EXPIRE_SECONDS)


class AltchaChallengeView(View):
    max_number = None
    expires = None

    @method_decorator(require_GET)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def get(self, request, *args, **kwargs):
        # Use view's class attributes or kwargs
        max_number = kwargs.get("max_number", self.max_number)
        expires = kwargs.get("expires", self.expires)

        challenge = get_altcha_challenge(max_number=max_number, expires=expires)
        return JsonResponse(challenge.__dict__)



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: MIT
# See https://github.com/aboutcode-org/django-altcha for support or download.
# See https://aboutcode.org for more information about AboutCode FOSS projects.
#

import json

from django.test import TestCase

from django_altcha import AltchaWidget


class DjangoAltchaWidgetTest(TestCase):
    def test_widget_initialization_with_default_options(self):
        widget = AltchaWidget(options=None)
        self.assertNotIn("challengeurl", widget.options)
        self.assertNotIn("challengejson", widget.options)
        self.assertNotIn("auto", widget.options)

    def test_widget_initialization_with_custom_options(self):
        options = {
            "auto": "onload",
            "delay": 500,
            "expire": 100000,
        }
        widget = AltchaWidget(options)
        self.assertEqual(widget.options["auto"], "onload")
        self.assertEqual(widget.options["delay"], 500)
        self.assertEqual(widget.options["expire"], 100000)

    def test_widget_generates_challengejson_if_no_challengeurl(self):
        widget = AltchaWidget(options={})  # Pass an empty dictionary
        context = widget.get_context(name="test", value=None, attrs={})
        altcha_options = context["widget"]["altcha_options"]
        challengejson = json.loads(altcha_options["challengejson"])
        self.assertEqual("SHA-256", challengejson["algorithm"])
        self.assertEqual(64, len(challengejson["challenge"]))
        self.assertIn("?expires=", challengejson.get("salt"))
